import csv
import sys
import urllib.request
import zipfile
from io import BytesIO, StringIO

ZIP_FILE_URL = "https://www.ecb.europa.eu/stats/eurofxref/eurofxref-hist.zip"
ZIP_FILE_ARCHIVE_NAME = 'eurofxref-hist.csv'
CSV_DATE_KEY = 'Date'

zipfile = zipfile.ZipFile(BytesIO(urllib.request.urlopen(ZIP_FILE_URL).read()))
input_stream = StringIO(zipfile.read(ZIP_FILE_ARCHIVE_NAME).decode('utf-8'))
csv_reader = csv.DictReader(input_stream, delimiter=',')

csv_writer = csv.writer(sys.stdout, delimiter="\t")

for row in csv_reader:
    for currency in row.keys():
        # all currency codes have length 3
        if len(currency) != 3:
            continue
        if row[currency] == 'N/A':
            continue
        csv_writer.writerow([currency, row[currency], row[CSV_DATE_KEY]])



"""Parallel creation of an attribute lookup table for another table (e.g. for auto-completion)"""

import re

import data_integration.config
import mara_db.postgresql
from data_integration.commands.sql import ExecuteSQL
from data_integration.pipelines import Pipeline, ParallelTask, Task
from mara_page import _


class CreateAttributesTable(ParallelTask):
    def __init__(self, id: str, source_schema_name: str, source_table_name: str,
                 db_alias: str = None,
                 attributes_table_suffix: str = '_attributes',
                 max_number_of_parallel_tasks: int = None) -> None:
        """
        Creates an indexed lookup table for providing fast auto-completion on the values of a table

        Given a table `foo.bar` like

        a | b | c
        --+---+---
        1 | x | i
        2 | y | i
        3 | x | i

        it creates a table foo.bar_attributes such as

        attribute | value | row_count
        ----------+-------+----------
        b         | x     | 2
        b         | y     | 1
        c         | i     | 3

        Args:
            id: The id of the task
            source_schema_name: The schema of the original table, e.g. 'foo'
            source_table_name: The name of the original table, e.g. 'bar'
            db_alias: The database alias for the source and attributes table
            attributes_table_suffix: This suffix will be appended to the source table name
            max_number_of_parallel_tasks: How many child tasks to run at most
        """
        super().__init__(id,
                         description=f'Creates an attributes lookup table on {source_schema_name}.{source_table_name}.',
                         max_number_of_parallel_tasks=max_number_of_parallel_tasks)

        self.source_schema_name = source_schema_name
        self.source_table_name = source_table_name
        self.attributes_table_suffix = attributes_table_suffix
        self.db_alias = db_alias or data_integration.config.default_db_alias()

    def add_parallel_tasks(self, sub_pipeline: Pipeline) -> None:
        attributes_table_name = f'{self.source_schema_name}.{self.source_table_name}{self.attributes_table_suffix}'
        sub_pipeline.add_initial(
            Task(id='create_table', description='Creates the attributes table',
                 commands=[
                     ExecuteSQL(sql_statement=f'''
DROP TABLE IF EXISTS {attributes_table_name};

CREATE TABLE {attributes_table_name} (
    attribute TEXT NOT NULL, 
    value     TEXT NOT NULL, 
    row_count BIGINT NOT NULL
) PARTITION BY LIST (attribute);
''')
                 ]))

        with mara_db.postgresql.postgres_cursor_context(self.db_alias) as cursor:  # type: psycopg2.extensions.cursor
            cursor.execute(f'''
WITH enums AS (
    SELECT DISTINCT
      typname,
      nspname
    FROM pg_type
      JOIN pg_enum ON pg_type.oid = pg_enum.enumtypid
      JOIN pg_namespace ON pg_type.typnamespace = pg_namespace.oid
  )
SELECT column_name
FROM information_schema.columns
  LEFT JOIN enums ON udt_schema = enums.nspname AND udt_name = enums.typname
  WHERE table_schema = {'%s'}
      AND table_name = {'%s'}
      AND (data_type IN ('text', 'varchar') OR enums.typname IS NOT NULL);
''', (self.source_schema_name, self.source_table_name))

            i = 0

            for column_name, in cursor.fetchall():
                i += 1
                sub_pipeline.add(Task(
                    id=re.sub('\W+', '_', column_name.lower()),
                    description=f'Extracts attributes for the {column_name} column',
                    commands=[
                        ExecuteSQL(sql_statement=f'''
CREATE TABLE {attributes_table_name}_{i} PARTITION OF {attributes_table_name} FOR VALUES IN ('{column_name}');

INSERT INTO {attributes_table_name}_{i} 
SELECT '{column_name}', "{column_name}", count(*)
FROM {self.source_schema_name}.{self.source_table_name}
WHERE "{column_name}" IS NOT NULL
GROUP BY "{column_name}"
ORDER BY "{column_name}";

CREATE INDEX {self.source_table_name}_{self.attributes_table_suffix}_{i}__value 
   ON {attributes_table_name}_{i} USING GIN (value gin_trgm_ops);
''', echo_queries=False)
                    ]))

    def html_doc_items(self) -> [(str, str)]:
        return [('db', _.tt[self.db_alias]),
                ('source schema', _.tt[self.source_schema_name]),
                ('source table', _.tt[self.source_table_name]),
                ('attributes table suffix', _.tt[self.attributes_table_suffix])]



import pathlib

from data_integration.commands.sql import ExecuteSQL
from data_integration.pipelines import Pipeline, Task
from etl_tools import config

pipeline = Pipeline(
    id="create_time_dimensions",
    description="Creates a day and a duration dimension table",
    labels={"Schema": "time"},
    base_path=pathlib.Path(__file__).parent)

pipeline.add(
    Task(id="create_tables",
         description="Re-creates the day and duration table and their schema",
         commands=[
             ExecuteSQL(sql_file_name='create_tables.sql', echo_queries=False,
                        file_dependencies=['create_tables.sql'])
         ]))

pipeline.add(
    Task(id="populate_time_dimensions", description="fills the time dimensions for a configured time range",
         commands=[
             ExecuteSQL(sql_statement=lambda: "SELECT time.populate_time_dimensions('"
                                              + config.first_date_in_time_dimensions().isoformat() + "'::DATE, '"
                                              + config.last_date_in_time_dimensions().isoformat() + "'::DATE);")]),
    upstreams=['create_tables'])



"""Configuration of shared ETL utilities"""

import datetime


def first_date_in_time_dimensions() -> datetime.date:
    """The first date that should appear in time dimensions"""
    return datetime.date.today() - datetime.timedelta(days=365)


def last_date_in_time_dimensions() -> datetime.date:
    """The last date that should appear in time dimensions"""
    return datetime.date.today()


def number_of_chunks() -> int:
    """Big tables and computations are split into this many chunks"""
    return 7



import pathlib
import os

from data_integration.commands.files import ReadScriptOutput
from data_integration.commands.sql import ExecuteSQL
from data_integration.pipelines import Pipeline, Task


def euro_exchange_rates_pipeline(db_alias: str):
    pipeline = Pipeline(
        id="load_euro_exchange_rates",
        description="Loads daily Euro exchange rates since 1999 from the European central bank",
        base_path=pathlib.Path(__file__).parent)

    pipeline.add(
        Task(id="create_schema_and_table",
             description="Re-creates currency exchange rate schema",
             commands=[
                 ExecuteSQL(sql_file_name='create_schema_and_table.sql', echo_queries=False)
             ]))

    pipeline.add(
        Task(id='load_exchange_rate', description='Loads exchange rates from the European central bank',
             commands=[ReadScriptOutput(file_name='load_exchange_rate.py', target_table='euro_fx.exchange_rate',
                       db_alias=db_alias)]),
        upstreams=['create_schema_and_table'])

    pipeline.add(
        Task(id="postprocess_exchange_rate",
             description="Adds values for missing days",
             commands=[
                 ExecuteSQL(sql_file_name='postprocess_exchange_rate.sql', echo_queries=False)
             ]),
        upstreams=['load_exchange_rate'])

    return pipeline



from etl_tools import config


def chunk_parameter_function() -> [(str)]:
    """Returns all chunks. Meant to be used in chunking-based parallel tasks"""
    return [(chunk,) for chunk in range(0, config.number_of_chunks())]


"""Machinery for copying whole schemas from one PostgreSQL database to another"""

import shlex

import mara_db.dbs
import mara_db.postgresql
import mara_db.shell
from data_integration.commands import bash
from data_integration.commands.bash import RunBash
from data_integration.commands.sql import ExecuteSQL
from data_integration.pipelines import Pipeline, Task, ParallelTask, Command
from mara_page import _


def add_schema_copying_to_pipeline(pipeline: Pipeline, schema_name,
                                   source_db_alias: str, target_db_alias: str,
                                   max_number_of_parallel_tasks: int = 4):
    """
    Adds schema copying to the end of a pipeline.

    When the pipeline already has a final node, then the all except the last command are run before the copying,
    and the last command after.

    Args:
        pipeline: The pipeline to modify
        schema_name: The schema to copy
        source_db_alias: The alias of the PostgreSQL database to copy from
        target_db_alias: The alias of the PostgreSQL database to copy to
        max_number_of_parallel_tasks: How many operations to run at parallel at max.
    """
    task_id = "copy_schema"
    description = f"Copies the {schema_name} schema to the {target_db_alias} db"
    commands = []
    if pipeline.final_node:
        assert (isinstance(pipeline.final_node, Task))
        description = pipeline.final_node.description + ' + ' + description
        task_id = pipeline.final_node.id + '_and_' + task_id
        commands = pipeline.final_node.commands
        pipeline.remove(pipeline.final_node)

    pipeline.add_final(
        ParallelCopySchema(id=task_id, description=description, schema_name=schema_name,
                           source_db_alias=source_db_alias, target_db_alias=target_db_alias,
                           max_number_of_parallel_tasks=max_number_of_parallel_tasks,
                           commands_before=commands[:-1], commands_after=commands[-1:]))


class ParallelCopySchema(ParallelTask):
    def __init__(self, id: str, description: str, max_number_of_parallel_tasks: int,
                 source_db_alias: str, target_db_alias: str, schema_name: str,
                 commands_before: [Command] = None, commands_after: [Command] = None) -> None:
        """In parallel copies a PostgreSQL database schema from one database to another."""

        ParallelTask.__init__(self, id=id, description=description,
                              max_number_of_parallel_tasks=max_number_of_parallel_tasks,
                              commands_before=commands_before, commands_after=commands_after)

        self.source_db_alias = source_db_alias
        self.target_db_alias = target_db_alias
        self.schema_name = schema_name

    def add_parallel_tasks(self, sub_pipeline: Pipeline) -> None:
        source_db = mara_db.dbs.db(self.source_db_alias)
        target_db = mara_db.dbs.db(self.target_db_alias)
        assert (isinstance(source_db, mara_db.dbs.PostgreSQLDB))
        assert (isinstance(target_db, mara_db.dbs.PostgreSQLDB))

        with mara_db.postgresql.postgres_cursor_context(self.source_db_alias) as cursor:
            pg_version = cursor.connection.server_version

        ddl_task = Task(
            id='create_tables_and_functions',
            description='Re-creates the schema, tables structure and functions on the target db',
            commands=[
                # schema and table structure
                bash.RunBash(
                    command="(echo 'DROP SCHEMA IF EXISTS " + self.schema_name + " CASCADE;';\\\n"
                            + "    pg_dump --username=" + source_db.user + " --host=" + source_db.host
                            + " --schema=" + self.schema_name
                            + " --section=pre-data --no-owner --no-privileges " + source_db.database + ") \\\n"
                            + "  | " + mara_db.shell.query_command(self.target_db_alias,
                                                                   echo_queries=False) + ' --quiet'),

                # function definitions
                bash.RunBash(
                    command=f'''echo "
SELECT CONCAT(pg_get_functiondef(pg_proc.oid),';') AS def 
FROM (SELECT oid, * 
      FROM pg_proc p 
      WHERE {"p.prokind in ('p','f')" if pg_version >= 110000 else "NOT p.proisagg"}) pg_proc, pg_namespace
WHERE pg_proc.pronamespace = pg_namespace.oid
     AND nspname = '{self.schema_name}'" \\\n'''
                            + "  | " + mara_db.shell.copy_to_stdout_command(self.source_db_alias) + ' \\\n'
                            + "  | " + mara_db.shell.query_command(self.target_db_alias, echo_queries=False))
            ])
        sub_pipeline.add(ddl_task)

        # copy content of tables
        number_of_chunks = self.max_number_of_parallel_tasks * 3
        table_copy_chunks = {i: [] for i in range(0, number_of_chunks)}
        current_size_per_table_copy_chunk = [0] * number_of_chunks
        table_types = {}

        with mara_db.postgresql.postgres_cursor_context(
                self.source_db_alias) as cursor:  # type: psycopg2.extensions.cursor
            cursor.execute("""
SELECT 
    pg_class.relname AS table,
    relkind,
    CASE WHEN relkind = 'f' 
         THEN cstore_table_size(nspname || '.' || relname) * 10 -- cstore tables with similar size take longer to copy 
         ELSE  pg_total_relation_size(pg_class.oid)
    END / 1000000.0 AS size
FROM pg_class
JOIN pg_namespace ON pg_namespace.oid = pg_class.relnamespace
WHERE nspname = '""" + self.schema_name + """' AND relkind IN ('r', 'f') AND relhassubclass = 'f'
ORDER BY size DESC""")
            for table_name, type, size in cursor.fetchall():
                smallest_chunk_index = min(range(len(current_size_per_table_copy_chunk)),
                                           key=current_size_per_table_copy_chunk.__getitem__)
                current_size_per_table_copy_chunk[smallest_chunk_index] += size
                table_copy_chunks[smallest_chunk_index].append(table_name)
                table_types[table_name] = type

            copy_tasks = []
            for i, tables in table_copy_chunks.items():
                if tables:
                    task = Task(
                        id=f'copy_tables_{i}',
                        description='Copies table content to the frontend db',
                        commands=[RunBash(
                            command=f'echo {shlex.quote(f"COPY {self.schema_name}.{table_name} TO STDOUT")} \\\n'
                                    + '  | ' + mara_db.shell.copy_to_stdout_command(self.source_db_alias) + ' \\\n'
                                    + '  | ' + mara_db.shell.copy_from_stdin_command(self.target_db_alias,
                                                                                     target_table=f'{self.schema_name}.{table_name}'))
                            for table_name in tables])
                    copy_tasks.append(task)
                    sub_pipeline.add(task, upstreams=[ddl_task])

            # create indexes
            index_chunks = {i: [] for i in range(0, number_of_chunks)}
            current_size_per_index_chunk = [0] * number_of_chunks

            with mara_db.postgresql.postgres_cursor_context(self.source_db_alias) as cursor:
                cursor.execute(""" 
SELECT indexdef AS ddl, pg_total_relation_size(pg_class.oid) AS size
FROM pg_class
JOIN pg_namespace ON pg_namespace.oid = pg_class.relnamespace
JOIN pg_indexes ON pg_indexes.indexname = pg_class.relname AND pg_indexes.schemaname = nspname
WHERE nspname = '""" + self.schema_name + """' AND relkind = 'i'
ORDER BY size DESC;""")
                for ddl, size in cursor.fetchall():
                    smallest_chunk_index = min(range(len(current_size_per_index_chunk)),
                                               key=current_size_per_index_chunk.__getitem__)
                    current_size_per_index_chunk[smallest_chunk_index] += size
                    index_chunks[smallest_chunk_index].append(ddl)

            for i, index_statements in index_chunks.items():
                if index_statements:
                    index_task = Task(id=f'add_indexes_{i}', description='Re-creates indexes on frontend db',
                                      commands=[ExecuteSQL(sql_statement=statement, db_alias=self.target_db_alias)
                                                for statement in index_statements])
                    sub_pipeline.add(index_task, upstreams=copy_tasks)

    def html_doc_items(self) -> [(str, str)]:
        return [('schema', _.tt[self.schema_name]),
                ('source db', _.tt[self.source_db_alias]),
                ('target db', _.tt[self.target_db_alias])]



import pathlib

from data_integration.commands.sql import ExecuteSQL
from data_integration.pipelines import Pipeline, Task
from etl_tools import config


def utils_pipeline(with_hll=False, with_cstore_fdw=False):
    pipeline = Pipeline(
        id="initialize_utils",
        description="Creates an utils schema with a number of functions around the ETL best practices of Project A",
        base_path=pathlib.Path(__file__).parent)

    pipeline.add_initial(
        Task(
            id="create_utils_schema",
            description="Re-creates the utils schema",
            commands=[
                ExecuteSQL(sql_statement="DROP SCHEMA IF EXISTS util CASCADE; CREATE SCHEMA util;")
            ]))

    pipeline.add(
        Task(id='chunking',
             description='Runs file chunking.sql',
             commands=[
                 ExecuteSQL(sql_file_name='chunking.sql', echo_queries=False,
                            replace={'number_of_chunks': lambda: config.number_of_chunks()})
             ]))

    def add_task_for_file(file_name_without_extension):
        pipeline.add(
            Task(id=file_name_without_extension,
                 description=f'Runs file "{file_name_without_extension}.sql"',
                 commands=[
                     ExecuteSQL(sql_file_name=file_name_without_extension + '.sql',
                                echo_queries=False)
                 ]))

    for file_name_without_extension in ['consistency_checks', 'data_sets', 'partitioning',
                                        'indexes_and_constraints',  'schema_switching', 'enums']:
        add_task_for_file(file_name_without_extension)

    if with_hll:
        add_task_for_file('hll')

    if with_cstore_fdw:
        add_task_for_file('cstore_fdw')

    return pipeline